//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2019 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package auth

import (
	"encoding/json"
	"io/ioutil"
	"log"
	"net/http"
	"strings"

	"github.com/gorilla/handlers"
	"github.com/gorilla/mux"
)

var auth = &Auth{}

type Auth struct {
	sessions *SessionManager
	oidc     *OIDCBackend
	infoLog  *log.Logger
	errLog   *log.Logger
	dbgLog   *log.Logger
}

func Init(c *Config, infoLog, errLog, dbgLog *log.Logger) (err error) {
	if infoLog == nil {
		infoLog = log.New(ioutil.Discard, "", 0)
	}
	if errLog == nil {
		errLog = log.New(ioutil.Discard, "", 0)
	}
	if dbgLog == nil {
		dbgLog = log.New(ioutil.Discard, "", 0)
	}

	if c == nil {
		infoLog.Println("authentication: disabled - will allow any request!")
		return
	}
	auth.infoLog = infoLog
	auth.errLog = errLog
	auth.dbgLog = dbgLog

	if auth.sessions, err = NewSessionManager(c.Sessions); err != nil {
		auth.errLog.Printf("authentication: failed to initialize session manager: %v", err)
		return
	}

	if c.OIDC != nil {
		if auth.oidc, err = NewOIDCBackend(c.OIDC); err != nil {
			auth.errLog.Printf("authentication: failed to initialize OIDC backend: %v", err)
			return
		}
		auth.infoLog.Printf("authentication: successfully initialized %s", auth.oidc)
	}

	auth.infoLog.Println("authentication: initialized")
	return
}

func newSession() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		request := &NewSessionRequest{}
		err := json.NewDecoder(r.Body).Decode(request)
		if err != nil {
			sendHTTPErrorResponse(w, http.StatusBadRequest, "Error decoding request: "+err.Error())
			return
		}

		response := &NewSessionResponse{}
		switch strings.ToLower(request.Backend) {
		case "oidc":
			if auth.oidc == nil {
				sendHTTPErrorResponse(w, http.StatusConflict, "OIDC authentication is not enabled")
				return
			}
			s, err := auth.oidc.NewOIDCSession(r.Context(), request.Arguments)
			if err != nil {
				sendHTTPErrorResponse(w, http.StatusBadRequest, "Error creating session: "+err.Error())
				return
			}
			response.Session = s
			response.Token = s.getToken()
		default:
			sendHTTPErrorResponse(w, http.StatusBadRequest, "invalid authentication backend: "+request.Backend)
			return
		}
		sendHTTPResponse(w, http.StatusOK, response)
	})
}

func getSession() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		s := getSessionFromBearerToken(r)
		if s == nil {
			sendHTTPInvalidSessionResponse(w)
			return
		}

		if r.URL.Query().Get("wait-for-login") == "" {
			sendHTTPResponse(w, http.StatusOK, s)
			return
		}

		ctx := r.Context()
		for {
			s, sub := auth.sessions.getAndSubscribe(s.ID())
			if s == nil {
				sendHTTPErrorResponse(w, http.StatusNotFound, "this session does no longer exist")
				return
			}
			st := s.State()
			if st >= SessionStateLoggedIn {
				sendHTTPResponse(w, http.StatusOK, s)
				return
			}
			select {
			case <-sub:
			case <-ctx.Done():
				sendHTTPErrorResponse(w, http.StatusRequestTimeout, "error while waiting for state change: "+ctx.Err().Error())
				return
			}
		}
	})
}

func deleteSession() http.Handler {
	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		s := getSessionFromBearerToken(r)
		if s == nil {
			sendHTTPInvalidSessionResponse(w)
			return
		}
		s.setState(SessionStateLoggedOut)
		sendHTTPResponse(w, http.StatusOK, "you are now logged out")
	})
}

func listBackends(w http.ResponseWriter, r *http.Request) {
	backends := []AuthBackendInfo{}
	if auth.oidc != nil {
		backend := AuthBackendInfo{Name: "oidc"}
		backend.Description = auth.oidc.String()
		backends = append(backends, backend)
	}

	sendHTTPResponse(w, http.StatusOK, backends)
}

func disabled(w http.ResponseWriter, r *http.Request) {
	sendHTTPErrorResponse(w, http.StatusBadRequest, "authentication is disabled")
}

func InstallHTTPHandler(r *mux.Router) {
	if auth.sessions == nil {
		r.HandleFunc("/backends", http.NotFound)
		r.PathPrefix("/").HandlerFunc(disabled)
		return
	}

	sessionHandler := make(handlers.MethodHandler)
	sessionHandler[http.MethodPost] = newSession()
	sessionHandler[http.MethodGet] = getSession()
	sessionHandler[http.MethodDelete] = deleteSession()
	r.Handle("/session", sessionHandler)

	r.HandleFunc("/backends", listBackends)
	if auth.oidc != nil {
		r.Handle("/oidc/login", auth.oidc.LoginHandler())
		r.Handle("/oidc/callback", auth.oidc.CallbackHandler())
	}
}

func Middleware(next http.Handler) http.Handler {
	if auth.sessions == nil {
		return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
			next.ServeHTTP(w, attachSessionToRequest(r, anonAllowAll))
		})
	}

	return http.HandlerFunc(func(w http.ResponseWriter, r *http.Request) {
		s := getSessionFromBearerToken(r)
		if s == nil || s.State() != SessionStateLoggedIn {
			sendHTTPInvalidSessionResponse(w)
			return
		}
		next.ServeHTTP(w, attachSessionToRequest(r, s))
	})
}
